import matplotlib
matplotlib.use('TkAgg') 
import scipy as sp
import numpy as np
import pylab as py
import time

AE=149.6e9
sun_mass=1.9891e30
earth_mass=5.974e24
earth_norm=earth_mass/sun_mass
G=6.67428e-11/(AE**3)*sun_mass
particlecounter=0


def close_all():
	py.close('all')
	
def phi_centrifugal(x,L):
	return -0.5*L**2.0*np.sum(x**2)
	
def phi_grav(x_test,m,x):
	return -G*np.sum(m/np.sqrt(np.sum((x-x_test)**2,1)))
	
def phi_coriolis(x_test,v_test,L):
	return np.sum(v_test*L*np.array([-x[1],x[0]]))
	
def phi_complete(x_test,x,m,L):
	return phi_centrifugal(x_test,L)+phi_grav(x_test,m,x)

def a_coriolis(v_test,L):
	return -2.0* L*np.array([-v_test[1],v_test[0]])
	
def a_g(x_test,x,m):
	distances=np.sqrt(np.sum((x-x_test)**2,1))
	dif=x-x_test
	
	return G*np.sum(((dif.T*m)/distances**3).T,0)

	
def a_centrifugal(x_test,L):
	return L**2*x_test
	
def a_complete(x_test,v_test,L,x,m):
	return a_coriolis(v_test,L)+a_g(x_test,x,m) + a_centrifugal(x_test,L)
	
def simulation(x_start,x,m,L,dt,tend,method):
	v_start=np.zeros(2)
	
	x_store=np.array([x_start.copy()])
	x_act=x_start.copy()
	v_act=v_start.copy()
	if method=='lf': leapfrog_init(x_act,v_start,x,m,L,dt)
	for t in np.arange(0,tend,dt):
		x_act,v_act=leap_frog(x_act,v_act,x,m,L,dt)
		x_store=np.append(x_store,np.array([x_act]),0)
		print t
		
	return x_store

def leap_frog(x_test,v_test,x,m,L,dt):
    a=a_complete(x_test,v_test,L,x,m)
    v_test=v_test+(a*dt)
    x_test=x_test+(v_test*dt)
    return x_test,v_test
    
def leapfrog_init(x_test,v_test,x,m,L,dt):
	print x_test+0.5*(v_test.T*dt).T+0.125*(a_complete(x_test,v_test,L,x,m).T*(dt**2)).T
	return x_test+0.5*(v_test.T*dt).T+0.125*(a_complete(x_test,v_test,L,x,m).T*(dt**2)).T

def RK(x_test,v_test,x,m,L,dt):
	x1=v_test.copy()
	v1=a_complete(x_test,v_test,L,x,m)

	x2=v_test+dt/2.*v1
	v2=a_complete(x_test+dt/2.0*x1,v_test,L,x,m)

	x3=v_test+dt/2.0*v2
	v3=a_complete(x_test+dt/2.0*x2,v_test,L,x,m)

	x4=v_test+dt*v3
	v4=a_complete(x_test+dt*x3,v_test,L,x,m)

	xnew=x_test+dt/6.*(x1+2.*x2+2.*x3+x4)
	vnew=v_test+dt/6.*(v1+2.*v2+2.*v3+v4)
	return xnew,vnew

def NR_step(f,x_in,h,x,m,L):
	H=Hessian(f,x_in,h,x,m,L)
	g=Gradient(f,x_in,h,x,m,L)
	return np.linalg.solve(H,-g)

def NR_full(f,x_in,h,x,m,L):
	x_act=x_in.copy()
	x_store=np.array([x_act])
	i=0
	while (1):
		i+=1		
		dx=NR_step(f,x_act,h,x,m,L)
		x_act+=dx*0.1
		print x_act
		x_store=np.append(x_store,np.array([x_act]),0)
		if np.linalg.norm(dx)<0.00000001:break
	return x_store
	
def Hessian(f,x_in,h,xp,m,L): ## Hessian at x_in (xp are the position of the 2 heavy masses)
	x=x_in.copy()
	N=np.size(x)
	H=np.zeros((N,N))
	for i in range(N):
		for j in range(i,N):
			x[i]+=h
			x[j]+=h
			H[i,j]+=f(x,xp,m,L)
			x[i]-=2.0*h
			H[i,j]-=f(x,xp,m,L)
			x[j]-=2.0*h
			H[i,j]+=f(x,xp,m,L)
			x[i]+=2.0*h
			H[i,j]-=f(x,xp,m,L)
			x[i]-=h
			x[j]+=h
	H/=4.0*h**2
	return H+H.T-np.diag(H)*np.eye(N)
	
def Gradient(f,x_in,h,xp,m,L):## Gradient of f at position x_in with discretization h(xp are the position of the 2 heavy masses)
	N=np.size(x_in)
	g=np.zeros(N)
	x=x_in.copy()
	print x
	for i in range(N):
		x[i]+=h
		g[i]+=f(x,xp,m,L)
		x[i]-=2.0*h
		g[i]-=f(x,xp,m,L)
		x[i]+=h
	return g/(2.0*h)
	
def main_iteration(m1,m2,R,method,dt):
	particlecounter = 0
	if m2>m1:
		zw=m1
		m1=m2
		m2=zw
	Omega=np.sqrt(G/R**3*(m1+m2)) ## angular velocity of the system --> from Kepler's laws
	r1=R*m2/(m1+m2)
	r2=R*m1/(m1+m2)
	x1=np.array([-r1,0.0])
	x2=np.array([r2,0.0])
	x=np.array([x1,x2])
	m=np.array([m1,m2])
	
	## simulation box size and discretization --> spatial discretization only needed for visualization
	box_width=R*1.5
	dx=0.025
	
	## calculation of the potential at every grid point in the box
	X,Y=np.meshgrid(np.arange(-box_width,box_width,dx),np.arange(-box_width,box_width,dx))
	x_grid=np.array([X,Y])
	x_grid=x_grid.transpose(2,1,0)
	N=int(np.floor(2.0*box_width/dx))
	phi=np.zeros((N,N))
	
	alpha=m[1]/(m[0]+m[1])
	L1=R*np.array([1.0-(alpha/3.0)**(1.0/3.0),0])
	L2=R*np.array([1.0+(alpha/3.0)**(1.0/3.0),0])
	L4=np.array([0.5*R*(m[0]-m[1])/(m[0]+m[1]),np.sqrt(3.0)/2.0*R])
	L5=L4*np.array([1,-1])
	phil1=phi_complete(L1,x,m,Omega)
	phil4=phi_complete(L4,x,m,Omega)
	span=phil4-phil1
	ratio = max(m1/m2,m2/m1)
	cutoff=phil4-0.4*np.sqrt(ratio)*span#phil4-1.5*span

	for i in range(N):
		for j in range(N):
			phi[i,j]=phi_complete(x_grid[i,j],x,m,Omega)
			if phi[i,j]<cutoff: phi[i,j]=cutoff
	
	## ifgure parameters
	fig1=py.figure(figsize=(10,8))
	ax1=py.subplot(111)
	print ax1.axes.get_position()
	ax1.axes.set_position([0.075,0.03,0.85,0.83])
	##title
	py.suptitle("Press right mouse button for Newton-Raphson search of the stationary states \n Press middle mouse button to place an object into the potential\n")
	ax1.set_autoscale_on(False)
	ax1.set_title('Domain of size ('+str(box_width)+' AU)$^2$')

	ax1.set_axis_off()
	matplotlib.rcParams['contour.negative_linestyle']='solid'
	
	py.contour(phi.T,15,colors='black',linestyles='dotted',linewidth=1)
	py.contourf(phi.T,60,cmap=py.cm.gray,linewidth=1)
	
	py.colorbar()

	py.axis('equal') ##plot gets the right aspect ratio of 1
	ax=py.axis()
	
	x_start=np.array([-0.5,0.0])
	
	can=fig1.canvas
	fr=fig1.canvas.get_tk_widget()
	t_konst=60*60*24.0
	
	if method =='lf': meth=leap_frog
	else: meth=RK
	
	def update(line,x_store,v_old,number):
		x_act,v_act=meth(x_store[-1],v_old,x,m,Omega,dt)
		run=True
		x_store=np.append(x_store,np.array([x_act]),0)
		line.set_xdata(x_store[:,0]/dx+N/2)
		line.set_ydata(x_store[:,1]/dx+N/2)
		if run: fr.after(1,update,line,x_store,v_act,number)
		if (np.size(x_store,0))%100==0:
			print 'test mass: ',number,' t: ',np.size(x_store,0)*dt/t_konst,' days'
			py.draw()
	
	def onclick(event):
		global particlecounter
		
		if event.button==1: return False
		
		ax=py.axis()
		x_start=np.array([(event.xdata-N/2)*dx,(event.ydata-N/2)*dx])
		v_start=np.zeros(2)
		if event.button==3:
			x_store=NR_full(phi_complete,x_start,0.00001,x,m,Omega)
			ax1.set_autoscale_on(False)
			py.plot(x_store[:,0]/dx+N/2,x_store[:,1]/dx+N/2,'red',lw=2)
		else:
			if event.button==2:
				particlecounter+=1
				x_store=np.array([x_start])
				x_act=x_start.copy()
				if method=='lf': x_act=leapfrog_init(x_start,v_start,x,m,Omega,dt)
				ax1.set_autoscale_on(False)
				x_store=np.append(x_store,np.array([x_act]),0)
				line,=py.plot(x_store[:,0]/dx+N/2,x_store[:,1]/dx+N/2,'b',linewidth=3)
				fr.after(1,update,line,x_store,v_start,particlecounter)
		py.axis(ax)
		py.draw()
		return True
		
	
	x_store=np.array([[]])
	cid=can.mpl_connect('button_press_event', onclick)
	py.show()
	can.mpl_disconnect(cid)
	
	
if __name__=='__main__':
	main_iteration(1.0,0.01,1.0,'rk',1.0*60*60*24)
	# earth_mass/sun_mass
	
	
	
	
	
